#include "common.h"

#include <algorithm>
#include <numeric>
#include <vector>
#include <iterator>

void htmlencode(std::string& data) {
	std::string tmp;
	tmp.reserve(data.size() + 20);

	for (size_t i = 0; i != data.size(); i++) {
		switch (data[i]) {
			case '&':
				tmp += "&amp;";
				break;
			case '\"':
				tmp += "&quot;";
				break;
			case '\'':
				tmp += "&apos;";
				break;
			case '<':
				tmp += "&lt;";
				break;
			case '>':
				tmp += "&gt;";
				break;
			default:
				tmp += data[i];
				break;
		}
	}

	data.swap(tmp);
}

static std::string string_mangle(const std::string& str) {
		std::string tmp;
		tmp.reserve(str.size());

		for (size_t i = 0; i != str.size(); i++) {
		    if (str[i] == ' ') continue;
		    if (std::isupper(str[i])) tmp.push_back(std::tolower(str[i]));
		    else  tmp.push_back(str[i]);
		}
		return tmp;
}

void show_histogram(const Caches& cc, std::string Cache::*ptr, const std::string& caption, bool html, bool sort_by_val) {
	std::map<std::string, int> histogram;

	for (auto el : cc)
		histogram[el.*ptr]++;

	show_histogram(histogram, caption, html, sort_by_val);
}

void show_histogram(const std::map<std::string, int>& data, const std::string& caption, bool html, bool sort_by_val) {
	int HIST_MAX = 20;

	std::vector<std::pair<std::string, int>> pairs;

	if (!sort_by_val) HIST_MAX = 1000;

	std::copy(data.begin(), data.end(), std::back_inserter(pairs));

	if (sort_by_val)
		sort(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.second > b.second; });
	else
		sort(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.first < b.first; });

	if (html) {
		int max;
		if (sort_by_val)
			max = pairs[0].second;
		else
			max = std::max_element(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.second < b.second; })->second;
		int i = 0;
		std::cout << "<h2>" << caption << "</h2>\n";
		std::cout << "<div class=\"histogram " << string_mangle(caption) << "\">\n";
		for (auto own : pairs) {
			htmlencode(own.first);
			if (own.first.empty()) own.first = "[unknown]";

			if (i < HIST_MAX)
				std::cout << "<div class=\"bar\" style=\"--percent: " << 100 * own.second / max << "%;\"><span class=\"text\">" << own.first << ": " << own.second << "</span></div>\n";
			else if (i == HIST_MAX) {
				std::cout << "</div>\n";
				std::cout << "<details class=\"histogram_others " << string_mangle(caption) << "\">\n<summary>See more</summary>\n<p>\n";
				std::cout << own.first << " (" << own.second << ")";
			}
			if (i > HIST_MAX)
				std::cout << ", " << own.first << " (" << own.second << ")";
			i++;
		}
		if (i <= HIST_MAX)
			std::cout << "</div>\n";
		else
			std::cout << "</p></details>\n";
	} else {
		std::cout << caption << '\n';
		for (auto own : pairs)
			std::cout << own.first << ": " << own.second << '\n';
	}
}

void show_nested_histogram(const Caches& cc, std::string Cache::*ptr, std::string Cache::*ptr2, const std::string& caption, bool html, bool sort_by_val) {
	std::map<std::string, int> histogram;
	std::vector<std::pair<std::string, int>> pairs;

	for (auto el : cc)
		histogram[el.*ptr]++;
	std::copy(histogram.begin(), histogram.end(), std::back_inserter(pairs));

	if (sort_by_val)
		sort(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.second > b.second; });
	else
		sort(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.first < b.first; });

	if (html) {
		int HIST_MAX = 20;
		int max;

		if (sort_by_val)
			max = pairs[0].second;
		else
			max = std::max_element(pairs.begin(), pairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.second < b.second; })->second;
		int i = 0;
		std::cout << "<h2>" << caption << "</h2>\n";
		std::cout << "<div class=\"histogram nested " << string_mangle(caption) << "\">\n";
		for (auto own : pairs) {
			htmlencode(own.first);
			if (own.first.empty()) own.first = "[unknown]";
			if (i < HIST_MAX) {
				std::map<std::string, int> subhistogram;
				std::vector<std::pair<std::string, int>> subpairs;
				for (auto el : cc)
					if (el.*ptr == own.first && !(el.*ptr2).empty()) //TODO problem with [unknown]
						subhistogram[el.*ptr2]++;
				if (!subhistogram.empty()) {
					std::cout << "<details class=\"nested\"><summary class=\"bar\" style=\"--percent: " << 100 * own.second / max << "%;\"><span class=\"text\">" << own.first << ": " << own.second << "</span></summary>\n";
					std::copy(subhistogram.begin(), subhistogram.end(), std::back_inserter(subpairs));
					if (sort_by_val)
						sort(subpairs.begin(), subpairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.second > b.second; });
					else
						sort(subpairs.begin(), subpairs.end(), [&](std::pair<std::string, int>& a, std::pair<std::string, int>& b) { return a.first < b.first; });

					std::cout << "<div class=\"histogram inside\">\n";
					for (auto own : subpairs) {
						htmlencode(own.first);
						if (own.first.empty()) own.first = "[unknown]";
						std::cout << "<div class=\"bar\" style=\"--percent: " << 100 * own.second / max << "%;\"><span class=\"text\">" << own.first << ": " << own.second << "</span></div>\n";
					}
					std::cout << "</div>\n";
					std::cout << "</details>\n";
				} else {
					std::cout << "<div class=\"bar\" style=\"--percent: " << 100 * own.second / max << "%;\"><span class=\"text\">" << own.first << ": " << own.second << "</span></div>\n";
				}
			}
			else if (i == HIST_MAX) {
				std::cout << "</div>\n";
				std::cout << "<details class=\"histogram_others " << string_mangle(caption) << "\">\n<summary>See more</summary>\n<p>\n";
				std::cout << own.first << " (" << own.second << ")";
			}
			if (i > HIST_MAX)
				std::cout << ", " << own.first << " (" << own.second << ")";
			i++;
		}
		if (i <= HIST_MAX)
			std::cout << "</div>\n";
		else
			std::cout << "</p></details>\n";
	} else { //TODO text-only version not implemented, displays only first-level list
		std::cout << caption << '\n';
		for (auto own : pairs)
			std::cout << own.first << ": " << own.second << '\n';
	}
}

static bool same_day(const std::tm& a, const std::tm& b) {
	return (a.tm_year == b.tm_year && a.tm_mon == b.tm_mon && a.tm_mday == b.tm_mday);
}

static void next_day(std::tm& a) {
	a.tm_mday++;
	mktime(&a);
}

int find_streak(const std::multimap<std::time_t, const Cache*>& cc, std::tm& start) {
	int max_str = 0;
	int cur_str = 0;

	std::tm cur_tm;
	cur_tm.tm_year = 0;
	std::tm cur_start_tm = cur_tm;
	start = cur_tm;

	for (auto i : cc) {
		if (same_day(cur_tm, *std::localtime(&i.first)))
			continue;
		next_day(cur_tm);
		if (same_day(cur_tm, *std::localtime(&i.first))) {
			cur_str++;
			continue;
		}
		if (max_str < cur_str) {
			max_str = cur_str;
			start = cur_start_tm;
		}
		cur_str = 1;
		cur_tm = cur_start_tm = *std::localtime(&i.first);
		if (max_str == 0) {
			max_str = 1;
			start = cur_tm;
		}
	}

	if (max_str < cur_str) {
		max_str = cur_str;
		start = cur_start_tm;
	}

	return max_str;
}

long int get_num(char c, char* opt)  {
	try {
		if (std::stoi(opt) > 0) {
			return std::stoi(opt);
		} else {
			std::cout << "Option \"-" << c << "\" requires a valid number as an argument\n";
			std::exit(EXIT_FAILURE);
		}
	}
	catch (...) {
		std::cout << "Option \"-" << c << "\" requires a valid number as an argument\n";
		std::exit(EXIT_FAILURE);
	}
}

void average_html(const Caches& cc, float Cache::*ptr, const std::string& caption) {
		std::cout << "<div class=\"basic_stats\">\n";
		std::cout << "Average " << caption << ": <span class=\"value\">" << average(cc, ptr) << "</span><br>\n";
		std::cout << "</div>\n";
}

void average_html(const Caches& cc, int Cache::*ptr, const std::string& caption) {
		std::cout << "<div class=\"basic_stats\">\n";
		std::cout << "Average " << caption << ": <span class=\"value\">" << average(cc, ptr) << "</span><br>\n";
		std::cout << "</div>\n";
}

float average(const Caches& cc, float Cache::*ptr) {
	return std::accumulate(cc.begin(), cc.end(), 0., [&](const float& a, const Cache& b) { return std::move(a) + b.*ptr; }) / cc.size();
}

float average(const Caches& cc, int Cache::*ptr) {
	return std::accumulate(cc.begin(), cc.end(), 0, [&](const float& a, const Cache& b) { return std::move(a) + b.*ptr; }) / cc.size();
}
